@@ -1,17 +1,17 @@
 /*
-    commands.c: Commands sent to the card
-    Copyright (C) 2003-2010   Ludovic Rousseau
-    Copyright (C) 2005 Martin Paljak
+	commands.c: Commands sent to the card
+	Copyright (C) 2003-2010   Ludovic Rousseau
+	Copyright (C) 2005 Martin Paljak
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+	This library is free software; you can redistribute it and/or
+	modify it under the terms of the GNU Lesser General Public
+	License as published by the Free Software Foundation; either
+	version 2.1 of the License, or (at your option) any later version.
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
+	This library is distributed in the hope that it will be useful,
+	but WITHOUT ANY WARRANTY; without even the implied warranty of
+	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+	Lesser General Public License for more details.
 
 	You should have received a copy of the GNU Lesser General Public License
 	along with this library; if not, write to the Free Software Foundation,
@@ -20,18 +20,11 @@
 
 #include <config.h>
 
-#ifdef HAVE_STRING_H
+#include <stdbool.h>
 #include <string.h>
-#endif
-#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
-#endif
-#ifdef HAVE_ERRNO_H
 #include <errno.h>
-#endif
-#ifdef HAVE_UNISTD_H
 #include <unistd.h>
-#endif
 
 #include <pcsclite.h>
 #include <ifdhandler.h>
@@ -59,12 +52,6 @@
 #define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
 #endif
 
-#ifndef BSWAP_16
-#define BSWAP_8(x)  ((x) & 0xff)
-#define BSWAP_16(x) ((BSWAP_8(x) << 8) | BSWAP_8((x) >> 8))
-#define BSWAP_32(x) ((BSWAP_16(x) << 16) | BSWAP_16((x) >> 16))
-#endif
-
 #define CHECK_STATUS(res) \
 	if (STATUS_NO_SUCH_DEVICE == res) \
 		return IFD_NO_SUCH_DEVICE; \
@@ -101,9 +88,11 @@ RESPONSECODE CmdPowerOn(unsigned int reader_index, unsigned int * nlength,
 	unsigned char buffer[], int voltage)
 {
 	unsigned char cmd[10];
+	unsigned char resp[10 + MAX_ATR_SIZE];
+	int bSeq;
 	status_t res;
-	int length, count = 1;
-	unsigned int atr_len;
+	int count = 1;
+	unsigned int atr_len, length;
 	int init_voltage;
 	RESPONSECODE return_value = IFD_SUCCESS;
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
@@ -188,9 +177,6 @@ RESPONSECODE CmdPowerOn(unsigned int reader_index, unsigned int * nlength,
 	}
 #endif
 
-	/* store length of buffer[] */
-	length = *nlength;
-
 	if ((ccid_descriptor->dwFeatures & CCID_CLASS_AUTO_VOLTAGE)
 		|| (ccid_descriptor->dwFeatures & CCID_CLASS_AUTO_ACTIVATION))
 		voltage = 0;	/* automatic voltage selection */
@@ -201,54 +187,56 @@ RESPONSECODE CmdPowerOn(unsigned int reader_index, unsigned int * nlength,
 check_again:
 		if ((1 == voltage) && !(bVoltageSupport & 1))
 		{
-			DEBUG_INFO1("5V requested but not support by reader");
+			DEBUG_INFO1("5V requested but not supported by reader");
 			voltage = 2;	/* 3V */
 		}
 
 		if ((2 == voltage) && !(bVoltageSupport & 2))
 		{
-			DEBUG_INFO1("3V requested but not support by reader");
+			DEBUG_INFO1("3V requested but not supported by reader");
 			voltage = 3;	/* 1.8V */
 		}
 
 		if ((3 == voltage) && !(bVoltageSupport & 4))
 		{
-			DEBUG_INFO1("1.8V requested but not support by reader");
+			DEBUG_INFO1("1.8V requested but not supported by reader");
 			voltage = 1;	/* 5V */
-			goto check_again;
+
+			/* do not (infinite) loop if bVoltageSupport == 0 */
+			if (bVoltageSupport)
+				goto check_again;
 		}
 	}
 	init_voltage = voltage;
 
 again:
+	bSeq = (*ccid_descriptor->pbSeq)++;
 	cmd[0] = 0x62; /* IccPowerOn */
 	cmd[1] = cmd[2] = cmd[3] = cmd[4] = 0;	/* dwLength */
 	cmd[5] = ccid_descriptor->bCurrentSlotIndex;	/* slot number */
-	cmd[6] = (*ccid_descriptor->pbSeq)++;
+	cmd[6] = bSeq;
 	cmd[7] = voltage;
 	cmd[8] = cmd[9] = 0; /* RFU */
 
 	res = WritePort(reader_index, sizeof(cmd), cmd);
 	CHECK_STATUS(res)
 
-	/* reset available buffer size */
-	/* needed if we go back after a switch to ISO mode */
-	*nlength = length;
+	length = sizeof resp;
 
-	res = ReadPort(reader_index, nlength, buffer);
+	res = ReadPort(reader_index, &length, resp, bSeq);
 	CHECK_STATUS(res)
 
-	if (*nlength < STATUS_OFFSET+1)
+	if (length < CCID_RESPONSE_HEADER_SIZE)
 	{
-		DEBUG_CRITICAL2("Not enough data received: %d bytes", *nlength);
+		DEBUG_CRITICAL2("Not enough data received: %d bytes", length);
 		return IFD_COMMUNICATION_ERROR;
 	}
 
-	if (buffer[STATUS_OFFSET] & CCID_COMMAND_FAILED)
+	if (resp[STATUS_OFFSET] & CCID_COMMAND_FAILED)
 	{
-		ccid_error(PCSC_LOG_ERROR, buffer[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);    /* bError */
+		ccid_error(PCSC_LOG_ERROR, resp[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);	/* bError */
 
-		if (0xBB == buffer[ERROR_OFFSET] &&	/* Protocol error in EMV mode */
+		if (0xBB == resp[ERROR_OFFSET] &&	/* Protocol error in EMV mode */
 			((GEMPC433 == ccid_descriptor->readerID)
 			|| (CHERRYXX33 == ccid_descriptor->readerID)))
 		{
@@ -271,7 +259,7 @@ again:
 		if (voltage)
 		{
 #ifndef NO_LOG
-			const char *voltage_code[] = { "auto", "5V", "3V", "1.8V" };
+			const char *voltage_code[] = { "1.8V", "5V", "3V", "1.8V" };
 #endif
 
 			DEBUG_INFO3("Power up with %s failed. Try with %s.",
@@ -291,13 +279,13 @@ again:
 	}
 
 	/* extract the ATR */
-	atr_len = dw2i(buffer, 1);	/* ATR length */
+	atr_len = dw2i(resp, 1);	/* ATR length */
 	if (atr_len > *nlength)
 		atr_len = *nlength;
-	else
-		*nlength = atr_len;
 
-	memmove(buffer, buffer+10, atr_len);
+	*nlength = atr_len;
+
+	memcpy(buffer, resp+10, atr_len);
 
 	return return_value;
 } /* CmdPowerOn */
@@ -320,6 +308,8 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 	RESPONSECODE ret;
 	status_t res;
 
+	uint32_t ulDataLength;
+
 	pvs = (PIN_VERIFY_STRUCTURE *)TxBuffer;
 	cmd[0] = 0x69;	/* Secure */
 	cmd[5] = ccid_descriptor->bCurrentSlotIndex;	/* slot number */
@@ -337,15 +327,16 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 
 	/* On little endian machines we are all set. */
 	/* If on big endian machine and caller is using host byte order */
-	if ((pvs->ulDataLength + 19  == TxLength) &&
-		(bei2i((unsigned char*)(&pvs->ulDataLength)) == pvs->ulDataLength))
+	ulDataLength = get_U32(&pvs->ulDataLength);
+	if ((ulDataLength + 19 == TxLength) &&
+		(bei2i((unsigned char*)(&pvs->ulDataLength)) == ulDataLength))
 	{
 		DEBUG_INFO1("Reversing order from big to little endian");
 		/* If ulDataLength is big endian, assume others are too */
 		/* reverse the byte order for 3 fields */
-		pvs->wPINMaxExtraDigit = BSWAP_16(pvs->wPINMaxExtraDigit);
-		pvs->wLangId = BSWAP_16(pvs->wLangId);
-		pvs->ulDataLength = BSWAP_32(pvs->ulDataLength);
+		p_bswap_16(&pvs->wPINMaxExtraDigit);
+		p_bswap_16(&pvs->wLangId);
+		p_bswap_32(&pvs->ulDataLength);
 	}
 	/* At this point we now have the above 3 variables in little endian */
 
@@ -359,7 +350,7 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 	 * The Cherry XX44 reader crashes with a wrong value */
 	if ((0x00 == TxBuffer[7]) || (TxBuffer[7] > 0x07))
 	{
-		DEBUG_INFO2("Correct bEntryValidationCondition (was 0x%02X)",
+		DEBUG_INFO2("Fix bEntryValidationCondition (was 0x%02X)",
 			TxBuffer[7]);
 		TxBuffer[7] = 0x02;
 	}
@@ -373,22 +364,34 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 		 * CCID message. The only value supported is 01h (display 1 message) */
 		if (0x01 != TxBuffer[8])
 		{
-			DEBUG_INFO2("Correct bNumberMessage for GemPC Pinpad (was %d)",
+			DEBUG_INFO2("Fix bNumberMessage for GemPC Pinpad (was %d)",
 				TxBuffer[8]);
 			TxBuffer[8] = 0x01;
 		}
 
 		/* The reader does not support, and actively reject, "max size reached"
-		 * and "timeout occured" validation conditions */
+		 * and "timeout occurred" validation conditions */
 		if (0x02 != TxBuffer[7])
 		{
-			DEBUG_INFO2("Correct bEntryValidationCondition for GemPC Pinpad (was %d)",
+			DEBUG_INFO2("Fix bEntryValidationCondition for GemPC Pinpad (was %d)",
 				TxBuffer[7]);
 			TxBuffer[7] = 0x02;	/* validation key pressed */
 		}
 
 	}
 
+	if ((VENDOR_GEMALTO == GET_VENDOR(ccid_descriptor->readerID))
+		&& (ccid_descriptor->gemalto_firmware_features))
+	{
+		int bEntryValidationCondition = ccid_descriptor->gemalto_firmware_features->bEntryValidationCondition;
+		if (TxBuffer[7] & ~bEntryValidationCondition)
+		{
+			DEBUG_INFO2("Fix bEntryValidationCondition (was 0x%02X)",
+				TxBuffer[7]);
+			TxBuffer[7] &= bEntryValidationCondition;
+		}
+	}
+
 	if ((DELLSCRK == ccid_descriptor->readerID)
 		|| (DELLSK == ccid_descriptor->readerID))
 	{
@@ -396,7 +399,7 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 		 * CCID message. The only value supported is 00h (no message) */
 		if (0x00 != TxBuffer[8])
 		{
-			DEBUG_INFO2("Correct bNumberMessage for Dell keyboard (was %d)",
+			DEBUG_INFO2("Fix bNumberMessage for Dell keyboard (was %d)",
 				TxBuffer[8]);
 			TxBuffer[8] = 0x00;
 		}
@@ -414,7 +417,7 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 		tmp = TxBuffer[6];
 		TxBuffer[6] = TxBuffer[5];
 		TxBuffer[5] = tmp;
-		DEBUG_INFO1("Correcting wPINMaxExtraDigit for Dell keyboard");
+		DEBUG_INFO1("Fix wPINMaxExtraDigit for Dell keyboard");
 	}
 #endif
 
@@ -547,12 +550,10 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 				unsigned char sblk[1]; /* we only need 1 byte of data */
 				t1_state_t *t1 = &get_ccid_slot(reader_index)->t1;
 				unsigned int slen;
-				int oldReadTimeout;
 
 				DEBUG_COMM2("CT sent S-block with wtx=%u", RxBuffer[DATA]);
 				t1->wtx = RxBuffer[DATA];
 
-				oldReadTimeout = ccid_descriptor->readTimeout;
 				if (t1->wtx > 1)
 				{
 					/* set the new temporary timeout at WTX card request */
@@ -571,16 +572,20 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 
 				ret = CCID_Transmit(t1 -> lun, slen, RxBuffer, 0, t1->wtx);
 				if (ret != IFD_SUCCESS)
-					return ret;
+					goto end;
 
 				/* I guess we have at least 6 bytes in RxBuffer */
 				*RxLength = 6;
 				ret = CCID_Receive(reader_index, RxLength, RxBuffer, NULL);
 				if (ret != IFD_SUCCESS)
-					return ret;
+					goto end;
+			}
 
-				/* Restore initial timeout */
-				ccid_descriptor->readTimeout = oldReadTimeout;
+			/* this should not happen. It will make coverity happy */
+			if (*RxLength < 4)
+			{
+				ret = IFD_COMMUNICATION_ERROR;
+				goto end;
 			}
 
 			/* get only the T=1 data */
@@ -590,7 +595,9 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 	}
 
 end:
+	/* Restore initial timeout */
 	ccid_descriptor -> readTimeout = old_read_timeout;
+
 	return ret;
 } /* SecurePINVerify */
 
@@ -653,6 +660,7 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 	int bNumberMessage = 0; /* for GemPC Pinpad */
 	int gemalto_modify_pin_bug;
 #endif
+	uint32_t ulDataLength;
 
 	pms = (PIN_MODIFY_STRUCTURE *)TxBuffer;
 	cmd[0] = 0x69;	/* Secure */
@@ -671,15 +679,16 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 
 	/* On little endian machines we are all set. */
 	/* If on big endian machine and caller is using host byte order */
-	if ((pms->ulDataLength + 24  == TxLength) &&
-		(bei2i((unsigned char*)(&pms->ulDataLength)) == pms->ulDataLength))
+	ulDataLength = get_U32(&pms->ulDataLength);
+	if ((ulDataLength + 24 == TxLength) &&
+		(bei2i((unsigned char*)(&pms->ulDataLength)) == ulDataLength))
 	{
 		DEBUG_INFO1("Reversing order from big to little endian");
 		/* If ulDataLength is big endian, assume others are too */
 		/* reverse the byte order for 3 fields */
-		pms->wPINMaxExtraDigit = BSWAP_16(pms->wPINMaxExtraDigit);
-		pms->wLangId = BSWAP_16(pms->wLangId);
-		pms->ulDataLength = BSWAP_32(pms->ulDataLength);
+		p_bswap_16(&pms->wPINMaxExtraDigit);
+		p_bswap_16(&pms->wLangId);
+		p_bswap_32(&pms->ulDataLength);
 	}
 	/* At this point we now have the above 3 variables in little endian */
 
@@ -702,7 +711,7 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 	 * The Cherry XX44 reader crashes with a wrong value */
 	if ((0x00 == TxBuffer[10]) || (TxBuffer[10] > 0x07))
 	{
-		DEBUG_INFO2("Correct bEntryValidationCondition (was 0x%02X)",
+		DEBUG_INFO2("Fix bEntryValidationCondition (was 0x%02X)",
 			TxBuffer[10]);
 		TxBuffer[10] = 0x02;
 	}
@@ -735,15 +744,27 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 		|| (VEGAALPHA == ccid_descriptor->readerID))
 	{
 		/* The reader does not support, and actively reject, "max size reached"
-		 * and "timeout occured" validation conditions */
+		 * and "timeout occurred" validation conditions */
 		if (0x02 != TxBuffer[10])
 		{
-			DEBUG_INFO2("Correct bEntryValidationCondition for GemPC Pinpad (was %d)",
+			DEBUG_INFO2("Fix bEntryValidationCondition for GemPC Pinpad (was %d)",
 				TxBuffer[10]);
 			TxBuffer[10] = 0x02;	/* validation key pressed */
 		}
 	}
 
+	if ((VENDOR_GEMALTO == GET_VENDOR(ccid_descriptor->readerID))
+		&& (ccid_descriptor->gemalto_firmware_features))
+	{
+		int bEntryValidationCondition = ccid_descriptor->gemalto_firmware_features->bEntryValidationCondition;
+		if (TxBuffer[10] & ~bEntryValidationCondition)
+		{
+			DEBUG_INFO2("Fix bEntryValidationCondition (was 0x%02X)",
+				TxBuffer[10]);
+			TxBuffer[10] &= bEntryValidationCondition;
+		}
+	}
+
 	gemalto_modify_pin_bug = has_gemalto_modify_pin_bug(ccid_descriptor);
 	if (gemalto_modify_pin_bug)
 	{
@@ -756,11 +777,23 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 		bNumberMessage = TxBuffer[11];
 		if (0x03 != TxBuffer[11])
 		{
-			DEBUG_INFO2("Correct bNumberMessage for GemPC Pinpad (was %d)",
+			DEBUG_INFO2("Fix bNumberMessage for GemPC Pinpad (was %d)",
 				TxBuffer[11]);
 			TxBuffer[11] = 0x03; /* 3 messages */
 		}
 	}
+
+	/* Bug workaround for Cherry KC 1000 SC */
+	if (CHERRY_KC1000SC == ccid_descriptor->readerID)
+	{
+		/* The reader rejects bNumberMessage 0x00 in a PIN modify
+		 * command. Change it to 0xff which is accepted. */
+		if (0x00 == TxBuffer[11])
+		{
+			DEBUG_INFO1("Fix bNumberMessage for Cherry KC 1000 SC (was 0)");
+			TxBuffer[11] = 0xff;
+		}
+	}
 #endif
 
 	/* T=1 Protocol Management for a TPDU reader */
@@ -893,7 +926,7 @@ RESPONSECODE CmdEscape(unsigned int reader_index,
 	unsigned char RxBuffer[], unsigned int *RxLength, unsigned int timeout)
 {
 	return CmdEscapeCheck(reader_index, TxBuffer, TxLength, RxBuffer, RxLength,
-		timeout, FALSE);
+		timeout, false);
 } /* CmdEscape */
 
 
@@ -905,9 +938,10 @@ RESPONSECODE CmdEscape(unsigned int reader_index,
 RESPONSECODE CmdEscapeCheck(unsigned int reader_index,
 	const unsigned char TxBuffer[], unsigned int TxLength,
 	unsigned char RxBuffer[], unsigned int *RxLength, unsigned int timeout,
-	int mayfail)
+	bool mayfail)
 {
 	unsigned char *cmd_in, *cmd_out;
+	int bSeq;
 	status_t res;
 	unsigned int length_in, length_out;
 	RESPONSECODE return_value = IFD_SUCCESS;
@@ -938,10 +972,11 @@ again:
 		goto end;
 	}
 
+	bSeq = (*ccid_descriptor->pbSeq)++;
 	cmd_in[0] = 0x6B; /* PC_to_RDR_Escape */
 	i2dw(length_in - 10, cmd_in+1);	/* dwLength */
 	cmd_in[5] = ccid_descriptor->bCurrentSlotIndex;	/* slot number */
-	cmd_in[6] = (*ccid_descriptor->pbSeq)++;
+	cmd_in[6] = bSeq;
 	cmd_in[7] = cmd_in[8] = cmd_in[9] = 0; /* RFU */
 
 	/* copy the command */
@@ -961,7 +996,7 @@ again:
 
 time_request:
 	length_out = 10 + *RxLength;
-	res = ReadPort(reader_index, &length_out, cmd_out);
+	res = ReadPort(reader_index, &length_out, cmd_out, bSeq);
 
 	/* replay the command if NAK
 	 * This (generally) happens only for the first command sent to the reader
@@ -983,7 +1018,7 @@ time_request:
 		goto end;
 	}
 
-	if (length_out < STATUS_OFFSET+1)
+	if (length_out < CCID_RESPONSE_HEADER_SIZE)
 	{
 		free(cmd_out);
 		DEBUG_CRITICAL2("Not enough data received: %d bytes", length_out);
@@ -1001,7 +1036,7 @@ time_request:
 	{
 		/* mayfail: the error may be expected and not fatal */
 		ccid_error(mayfail ? PCSC_LOG_INFO : PCSC_LOG_ERROR,
-			cmd_out[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);    /* bError */
+			cmd_out[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);	/* bError */
 		return_value = IFD_COMMUNICATION_ERROR;
 	}
 
@@ -1033,6 +1068,7 @@ end:
 RESPONSECODE CmdPowerOff(unsigned int reader_index)
 {
 	unsigned char cmd[10];
+	int bSeq;
 	status_t res;
 	unsigned int length;
 	RESPONSECODE return_value = IFD_SUCCESS;
@@ -1085,20 +1121,21 @@ RESPONSECODE CmdPowerOff(unsigned int reader_index)
 	}
 #endif
 
+	bSeq = (*ccid_descriptor->pbSeq)++;
 	cmd[0] = 0x63; /* IccPowerOff */
 	cmd[1] = cmd[2] = cmd[3] = cmd[4] = 0;	/* dwLength */
 	cmd[5] = ccid_descriptor->bCurrentSlotIndex;	/* slot number */
-	cmd[6] = (*ccid_descriptor->pbSeq)++;
+	cmd[6] = bSeq;
 	cmd[7] = cmd[8] = cmd[9] = 0; /* RFU */
 
 	res = WritePort(reader_index, sizeof(cmd), cmd);
 	CHECK_STATUS(res)
 
 	length = sizeof(cmd);
-	res = ReadPort(reader_index, &length, cmd);
+	res = ReadPort(reader_index, &length, cmd, bSeq);
 	CHECK_STATUS(res)
 
-	if (length < STATUS_OFFSET+1)
+	if (length < CCID_RESPONSE_HEADER_SIZE)
 	{
 		DEBUG_CRITICAL2("Not enough data received: %d bytes", length);
 		return IFD_COMMUNICATION_ERROR;
@@ -1106,7 +1143,7 @@ RESPONSECODE CmdPowerOff(unsigned int reader_index)
 
 	if (cmd[STATUS_OFFSET] & CCID_COMMAND_FAILED)
 	{
-		ccid_error(PCSC_LOG_ERROR, cmd[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);    /* bError */
+		ccid_error(PCSC_LOG_ERROR, cmd[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);	/* bError */
 		return_value = IFD_COMMUNICATION_ERROR;
 	}
 
@@ -1122,6 +1159,7 @@ RESPONSECODE CmdPowerOff(unsigned int reader_index)
 RESPONSECODE CmdGetSlotStatus(unsigned int reader_index, unsigned char buffer[])
 {
 	unsigned char cmd[10];
+	int bSeq;
 	status_t res;
 	unsigned int length;
 	RESPONSECODE return_value = IFD_SUCCESS;
@@ -1171,7 +1209,7 @@ again_status:
 	if (PROTOCOL_ICCD_B == ccid_descriptor->bInterfaceProtocol)
 	{
 		int r;
-		unsigned char buffer_tmp[3];
+		unsigned char buffer_tmp[3] = {0, 2, 0};
 
 		/* SlotStatus */
 		r = ControlUSB(reader_index, 0xA1, 0x81, 0, buffer_tmp,
@@ -1201,27 +1239,29 @@ again_status:
 		}
 		return IFD_SUCCESS;
 	}
-#endif
 
 #ifdef __APPLE__
 	if (MICROCHIP_SEC1100 == ccid_descriptor->readerID)
 		InterruptRead(reader_index, 10);
 #endif
 
+#endif
+
+	bSeq = (*ccid_descriptor->pbSeq)++;
 	cmd[0] = 0x65; /* GetSlotStatus */
 	cmd[1] = cmd[2] = cmd[3] = cmd[4] = 0;	/* dwLength */
 	cmd[5] = ccid_descriptor->bCurrentSlotIndex;	/* slot number */
-	cmd[6] = (*ccid_descriptor->pbSeq)++;
+	cmd[6] = bSeq;
 	cmd[7] = cmd[8] = cmd[9] = 0; /* RFU */
 
 	res = WritePort(reader_index, sizeof(cmd), cmd);
 	CHECK_STATUS(res)
 
 	length = SIZE_GET_SLOT_STATUS;
-	res = ReadPort(reader_index, &length, buffer);
+	res = ReadPort(reader_index, &length, buffer, bSeq);
 	CHECK_STATUS(res)
 
-	if (length < STATUS_OFFSET+1)
+	if (length < CCID_RESPONSE_HEADER_SIZE)
 	{
 		DEBUG_CRITICAL2("Not enough data received: %d bytes", length);
 		return IFD_COMMUNICATION_ERROR;
@@ -1232,7 +1272,7 @@ again_status:
 		&& (buffer[ERROR_OFFSET] != 0xFE))
 	{
 		return_value = IFD_COMMUNICATION_ERROR;
-		ccid_error(PCSC_LOG_ERROR, buffer[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);    /* bError */
+		ccid_error(PCSC_LOG_ERROR, buffer[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);	/* bError */
 	}
 
 	return return_value;
@@ -1359,7 +1399,8 @@ RESPONSECODE CCID_Transmit(unsigned int reader_index, unsigned int tx_length,
 	cmd[8] = rx_length & 0xFF;	/* Expected length, in character mode only */
 	cmd[9] = (rx_length >> 8) & 0xFF;
 
-	memcpy(cmd+10, tx_buffer, tx_length);
+	if (tx_buffer)
+		memcpy(cmd+10, tx_buffer, tx_length);
 
 	ret = WritePort(reader_index, 10+tx_length, cmd);
 	CHECK_STATUS(ret)
@@ -1506,13 +1547,13 @@ time_request_ICCD_B:
 
 time_request:
 	length = sizeof(cmd);
-	ret = ReadPort(reader_index, &length, cmd);
+	ret = ReadPort(reader_index, &length, cmd, -1);
 
 	/* restore the original value of read timeout */
 	ccid_descriptor -> readTimeout = old_timeout;
 	CHECK_STATUS(ret)
 
-	if (length < STATUS_OFFSET+1)
+	if (length < CCID_RESPONSE_HEADER_SIZE)
 	{
 		DEBUG_CRITICAL2("Not enough data received: %d bytes", length);
 		return IFD_COMMUNICATION_ERROR;
@@ -1520,7 +1561,7 @@ time_request:
 
 	if (cmd[STATUS_OFFSET] & CCID_COMMAND_FAILED)
 	{
-		ccid_error(PCSC_LOG_ERROR, cmd[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);    /* bError */
+		ccid_error(PCSC_LOG_ERROR, cmd[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);	/* bError */
 		switch (cmd[ERROR_OFFSET])
 		{
 			case 0xEF:	/* cancel */
@@ -1542,6 +1583,12 @@ time_request:
 			case 0xFD:	/* Parity error during exchange */
 				return IFD_PARITY_ERROR;
 
+			case 0xFE:	/* Card absent or mute */
+				if (2 == (cmd[STATUS_OFFSET] & 0x02)) /* No ICC */
+					return IFD_ICC_NOT_PRESENT;
+				else
+					return IFD_COMMUNICATION_ERROR;
+
 			default:
 				return IFD_COMMUNICATION_ERROR;
 		}
@@ -1584,7 +1631,8 @@ time_request:
 		return_value = IFD_COMMUNICATION_ERROR;
 	}
 	else
-		memcpy(rx_buffer, cmd+10, length);
+		if (length)
+			memcpy(rx_buffer, cmd+10, length);
 
 	/* Extended case?
 	 * Only valid for RDR_to_PC_DataBlock frames */
@@ -1841,6 +1889,9 @@ static RESPONSECODE T0ProcACK(unsigned int reader_index,
 
 	DEBUG_COMM2("Enter, is_rcv = %d", is_rcv);
 
+	if (proc_len > 0x200)
+		return IFD_COMMUNICATION_ERROR;
+
 	if (is_rcv == 1)
 	{	/* Receiving mode */
 		unsigned int remain_len;
@@ -1886,29 +1937,29 @@ static RESPONSECODE T0ProcACK(unsigned int reader_index,
 
 #ifdef O2MICRO_OZ776_PATCH
 		if((0 != remain_len) && (0 == (remain_len + 10) % 64))
-        {
+		{
 			/* special hack to avoid a command of size modulo 64
 			 * we send two commands instead */
-            ret_len = 1;
-            return_value = CCID_Transmit(reader_index, 0, *snd_buf, ret_len, 0);
-            if (return_value != IFD_SUCCESS)
-                return return_value;
-            return_value = CCID_Receive(reader_index, &ret_len, tmp_buf, NULL);
-            if (return_value != IFD_SUCCESS)
-                return return_value;
+			ret_len = 1;
+			return_value = CCID_Transmit(reader_index, 0, *snd_buf, ret_len, 0);
+			if (return_value != IFD_SUCCESS)
+				return return_value;
+			return_value = CCID_Receive(reader_index, &ret_len, tmp_buf, NULL);
+			if (return_value != IFD_SUCCESS)
+				return return_value;
 
-            ret_len = remain_len - 1;
-            return_value = CCID_Transmit(reader_index, 0, *snd_buf, ret_len, 0);
-            if (return_value != IFD_SUCCESS)
-                return return_value;
-            return_value = CCID_Receive(reader_index, &ret_len, &tmp_buf[1],
-				NULL);
-            if (return_value != IFD_SUCCESS)
-                return return_value;
+			ret_len = remain_len - 1;
+			return_value = CCID_Transmit(reader_index, 0, *snd_buf, ret_len, 0);
+			if (return_value != IFD_SUCCESS)
+				return return_value;
+			return_value = CCID_Receive(reader_index, &ret_len, &tmp_buf[1],
+					NULL);
+			if (return_value != IFD_SUCCESS)
+				return return_value;
 
-            ret_len += 1;
-        }
-        else
+			ret_len += 1;
+		}
+		else
 #endif
 		{
 			ret_len = remain_len;
@@ -1937,6 +1988,12 @@ static RESPONSECODE T0ProcACK(unsigned int reader_index,
 		if (return_value != IFD_SUCCESS)
 			return return_value;
 
+		if (proc_len > *snd_len)
+		{
+			DEBUG_CRITICAL("proc_len > snd_len");
+			return IFD_COMMUNICATION_ERROR;
+		}
+
 		*snd_len -= proc_len;
 		*snd_buf += proc_len;
 	}
@@ -1983,10 +2040,14 @@ static RESPONSECODE T0ProcSW1(unsigned int reader_index,
 		in_buf = tmp_buf;
 	}
 	sw2 = *rcv_buf = *in_buf;
-	in_len--;
 	(*rcv_len)++;
 
+#ifdef NO_LOG
+	(void)sw1;
+	(void)sw2;
+#else
 	DEBUG_COMM3("Exit: SW=%02X %02X", sw1, sw2);
+#endif
 
 	return return_value;
 } /* T0ProcSW1 */
@@ -2078,7 +2139,7 @@ static RESPONSECODE CmdXfrBlockCHAR_T0(unsigned int reader_index,
 		if (0x20 == pcbuffer[0])
 		{
 			/* backup apdu data length */
-			/* if no data recieved before - backup length must be zero */
+			/* if no data received before - backup length must be zero */
 			backup_len = (backup_len == *rcv_len) ? 0 : *rcv_len;
 
 			/* wait for 2 bytes (SW1-SW2) */
@@ -2089,7 +2150,7 @@ static RESPONSECODE CmdXfrBlockCHAR_T0(unsigned int reader_index,
 			if (return_value != IFD_SUCCESS)
 				DEBUG_CRITICAL("CCID_Receive failed");
 
-			/* restore recieved length */
+			/* restore received length */
 			*rcv_len += backup_len;
 		}
 		return return_value;
@@ -2227,7 +2288,8 @@ static RESPONSECODE CmdXfrBlockTPDU_T1(unsigned int reader_index,
 
 	DEBUG_COMM3("T=1: %d and %d bytes", tx_length, *rx_length);
 
-	ret = t1_transceive(&((get_ccid_slot(reader_index)) -> t1), 0,
+	ret = t1_transceive(&((get_ccid_slot(reader_index)) -> t1),
+		get_ccid_slot(reader_index) -> t1.nad,
 		tx_buffer, tx_length, rx_buffer, *rx_length);
 
 	if (ret < 0)
@@ -2248,15 +2310,17 @@ RESPONSECODE SetParameters(unsigned int reader_index, char protocol,
 	unsigned int length, unsigned char buffer[])
 {
 	unsigned char cmd[10+length];	/* CCID + APDU buffer */
+	int bSeq;
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
 	status_t res;
 
 	DEBUG_COMM2("length: %d bytes", length);
 
+	bSeq = (*ccid_descriptor->pbSeq)++;
 	cmd[0] = 0x61; /* SetParameters */
 	i2dw(length, cmd+1);	/* APDU length */
 	cmd[5] = ccid_descriptor->bCurrentSlotIndex;	/* slot number */
-	cmd[6] = (*ccid_descriptor->pbSeq)++;
+	cmd[6] = bSeq;
 	cmd[7] = protocol;	/* bProtocolNum */
 	cmd[8] = cmd[9] = 0; /* RFU */
 
@@ -2266,10 +2330,10 @@ RESPONSECODE SetParameters(unsigned int reader_index, char protocol,
 	CHECK_STATUS(res)
 
 	length = sizeof(cmd);
-	res = ReadPort(reader_index, &length, cmd);
+	res = ReadPort(reader_index, &length, cmd, bSeq);
 	CHECK_STATUS(res)
 
-	if (length < STATUS_OFFSET+1)
+	if (length < CCID_RESPONSE_HEADER_SIZE)
 	{
 		DEBUG_CRITICAL2("Not enough data received: %d bytes", length);
 		return IFD_COMMUNICATION_ERROR;
@@ -2277,7 +2341,7 @@ RESPONSECODE SetParameters(unsigned int reader_index, char protocol,
 
 	if (cmd[STATUS_OFFSET] & CCID_COMMAND_FAILED)
 	{
-		ccid_error(PCSC_LOG_ERROR, cmd[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);    /* bError */
+		ccid_error(PCSC_LOG_ERROR, cmd[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);	/* bError */
 		if (0x00 == cmd[ERROR_OFFSET])	/* command not supported */
 			return IFD_NOT_SUPPORTED;
 		else
@@ -2318,7 +2382,7 @@ static void i2dw(int value, unsigned char buffer[])
 
 /*****************************************************************************
 *
-*                  bei2i (big endian integer to host order interger)
+*					bei2i (big endian integer to host order integer)
 *
 ****************************************************************************/
 
